import warnings

import numpy as np
import pandas as pd
import scipy as sp
from sklearn.exceptions import ConvergenceWarning, DataConversionWarning
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_val_score
from tqdm import tqdm

warnings.filterwarnings(action='ignore', category=ConvergenceWarning)
warnings.filterwarnings(action='ignore', category=DataConversionWarning)

def score(features, labels, setting='fast', cv=5):
    def wrapper(choice):
        if setting=='fast':
            clf = LogisticRegression(solver='lbfgs', multi_class='multinomial', max_iter=10)
        if setting=='slow':
            clf = LogisticRegression(solver='lbfgs', multi_class='multinomial', max_iter=100)
        return np.mean(cross_val_score(clf, features[:,choice], labels, cv=5))
    return wrapper

def classify(features, training_choice, training_labels):
    clf = LogisticRegression(solver='lbfgs', multi_class='multinomial', max_iter=100)
    clf.fit(training_choice, training_labels)
    return clf.predict(features)

def default_logger(*_):
    pass

def run(
        features, labels,
        num_mc_steps=10000,
        cluster_size=20,
        num_clusters=40,
        max_swaps_per_step=20,
        num_fast_mc_steps=20,
        log_noise=0.03,
        random_seed=0,
        logger=default_logger
):
    """
    Given input set of features and labels
    returns a subset of features that are optimal for classification
    """
    # setup
    np.random.seed(random_seed)

    features = np.array(features)
    labels = np.array(labels)
    score_fast = score(features, labels, setting='fast')
    score_slow = score(features, labels, setting='slow')

    num_features = features.shape[1]

    current_choice=np.random.choice(
        np.arange(num_features),
        size=cluster_size,
        replace=False
    )
    current_score=score_fast(current_choice)

    proposed_choice=current_choice
    proposed_score=current_score

    local_optimum_choice=current_choice
    local_optimum_score=current_score
    local_optimum_score_slow=score_slow(local_optimum_choice)

    global_optimum_choice=np.array([current_choice for _ in range(num_clusters)])
    global_optimum_score=np.array([local_optimum_score_slow for _ in range(num_clusters)])

    # outer Monte Carlo loop
    print('Running fastemc...')
    for _ in tqdm(range(num_mc_steps)):
        # inner "fast" Monte Carlo loop
        for _ in tqdm(range(num_fast_mc_steps)):
            num_to_swap = min(np.random.randint(2, max_swaps_per_step+1), num_features)
            proposed_insertion = np.random.choice(
                np.delete(np.arange(num_features), current_choice),
                size=num_to_swap,
                replace=False
            )
            proposed_removal = np.random.choice(
                cluster_size, 
                size=num_to_swap,
                replace=False
            )
            proposed_choice = current_choice
            np.put(proposed_choice, proposed_removal, proposed_insertion)

            proposed_score = score_fast(proposed_choice)

            if proposed_score > current_score - log_noise*np.log(1/np.random.rand()):
                current_choice = proposed_choice
                current_score = proposed_score

                if current_score > local_optimum_score:
                    local_optimum_choice = current_choice
                    local_optimum_score = current_score

        # slow scoring
        local_optimum_score_slow = score_slow(local_optimum_choice)
        if local_optimum_score_slow > np.min(global_optimum_score):
            min_index = np.argmin(global_optimum_score)
            global_optimum_choice[min_index] = local_optimum_choice
            global_optimum_score[min_index] = local_optimum_score_slow

        local_optimum_score = current_score
        local_optimum_choice = current_choice

        logger(global_optimum_score, global_optimum_choice)

    return global_optimum_score, global_optimum_choice



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fastemc",
    version="0.0.1",
    author="James Rowland",
    author_email="rowland.208@gmail.com",
    description="FastEMC is a method for dimensionality reduction.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rowland-208/fastemc",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
)


import argparse
import os

import numpy as np
import pandas as pd

import fastemc

parser = argparse.ArgumentParser(
    description='Fast Exponential Monte Carlo'
)

parser.add_argument(
    '--features', '-f', type=str, default='features.csv',
    help='input csv filename containing feature data'
)
parser.add_argument(
    '--labels', '-l', type=str, default='labels.csv',
    help='input csv filename containing class labels'
)
parser.add_argument(
    '--results', '-r', type=str, default='results.csv',
    help='results csv filename'
)
parser.add_argument(
    '--num_mc_steps', '-N', type=int, default=10000,
    help='number of features in each cluster used for classification'
)
parser.add_argument(
    '--cluster_size', '-c', type=int, default=20,
    help='number of features in each cluster used for classification'
)
parser.add_argument(
    '--num_clusters', '-k', type=int, default=40,
    help='number of features in each cluster used for classification'
)
parser.add_argument(
    '--max_swaps_per_step', '-m', type=int, default=20,
    help='number of features in each cluster used for classification'
)
parser.add_argument(
    '--num_fast_mc_steps', '-n', type=int, default=20,
    help='number of features in each cluster used for classification'
)
parser.add_argument(
    '--log_noise', '-T', type=float, default=0.03,
    help='number of features in each cluster used for classification'
)

args = parser.parse_args()

args.features = os.path.realpath(args.features)
args.labels = os.path.realpath(args.labels)
args.results = os.path.realpath(args.results)
def path_check(path):
    return (
        os.path.exists(path), 
        """
        File not found: {}.
        """.format(path)
    )
assert path_check(args.features)
assert path_check(args.labels)
assert path_check(args.results)

try:
    features = pd.read_csv(args.features, index_col=0)
except FileNotFoundError:
    print('Provide a features file using the --features flag')
try:
    labels = pd.read_csv(args.labels, index_col=0)
except FileNotFoundError:
    print('Provide a labels file using the --features flag')

def save_results(global_optimum_score, global_optimum_choice):
    ordering = np.argsort(global_optimum_score)
    global_optimum_score = global_optimum_score[ordering]
    global_optimum_choice = global_optimum_choice[ordering]
    results = pd.DataFrame(
        [
            [score, *features.columns[choice]]
            for score, choice in zip(global_optimum_score, global_optimum_choice)
        ],
        columns=[
            'score', 
            *('feature_{}'.format(i) for i in range(args.cluster_size))
        ]
    )
    results.to_csv(args.results)

_ = fastemc.run(
    features.values, labels.values,
    num_mc_steps=args.num_mc_steps,
    cluster_size=args.cluster_size,
    num_clusters=args.num_clusters,
    max_swaps_per_step=args.max_swaps_per_step,
    num_fast_mc_steps=args.num_fast_mc_steps,
    log_noise=args.log_noise,
    logger=save_results
)